"""
computer_graphics_demo.fern
===========================

Draw a fern.
"""
import taichi as ti
import numpy as np
from .utils import setpixel
from typing import List, Tuple

# https://en.wikipedia.org/wiki/Barnsley_fern
probability = [0.01, 0.07, 0.07, 0.85]
# means x[n + 1] = np.dot(pos, eq)
# where pos = [x[n], y[n], 1]
# y[n + 1] ditto
eqs = [
    np.array([[0, 0, 0], [0, 0.16, 0]]),
    np.array([[0.2, -0.26, 0], [0.23, 0.22, 1.6]]),
    np.array([[-0.15, 0.28, 0], [0.26, 0.24, 0.44]]),
    np.array([[0.85, 0.04, 0], [-0.04, 0.85, 1.6]]),
]


def fern(
    probability: List[float],
    eqs: List[np.ndarray],
    init: List[float],
    iteration_max: int,
) -> Tuple[List[float], List[float]]:
    """fern.

    :param probability:
    :type probability: List[float]
    :param eqs:
    :type eqs: List[np.ndarray]
    :param init:
    :type init: List[float]
    :param iteration_max:
    :type iteration_max: int
    :rtype: Tuple[List[float], List[float]]
    """
    pos = np.ones(3, dtype=float)
    pos[: len(init)] = init
    probability = np.add.accumulate(probability)  # type: ignore
    rands = np.random.rand(iteration_max)
    select = np.zeros(iteration_max, dtype=int)
    for i, p in enumerate(probability):
        select[rands < p] = i
        # mark these indice has been used
        rands[rands < p] = 1

    xs, ys = [], []
    for i in range(iteration_max):
        tmp = np.dot(eqs[select[i]], pos)
        pos[:2] = tmp
        xs += [tmp[0]]
        ys += [tmp[1]]

    return xs, ys


def paint(
    R: int, G: int, B: int, scale: int, iteration_max: int, x0: int, y0: int
) -> None:
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param scale:
    :type scale: int
    :param iteration_max:
    :type iteration_max: int
    :param x0:
    :type x0: int
    :param y0:
    :type y0: int
    :rtype: None
    """
    xs, ys = fern(probability, eqs, [0] * 2, iteration_max)
    for x, y in zip(xs, ys):
        x1, y1 = x0 + ti.floor(x * scale), y0 + ti.floor(y * scale)
        setpixel(x1, y1, R, G, B)



#! /usr/bin/env python
# Docstring {{{ #
"""Computer graphics algorithms demonstration.

usage: {binname} [-hVdns] [ [-v] | [-q|-qq] ] [-x <height>] [-y <width>]
    [-o <out>] [-c <color>] [-a <args>] [-t <time>] <command>

options:
    -h, --help              Show this screen.
    -V, --version           Show version.
    -d, --debug             Debug this program.
    -n, --dry-run           Don't show any canvas.
    -s, --second            Use second algorithm. See section algorithm.
    -v, --verbose           Output verbosity. (e.g., default arguments)
    -q, --quiet             Redirect stdout to null. Again for stderr.
    -x, --height <height>   Height of canvas. [default: 512]
    -y, --width <width>     Width of canvas. [default: 1024]
    -o, --output <out>      Save output to a file. See section output.
    -c, --color <color>     Color of graph. [default: 255,0,0]
    -a, --args <args>       Arguments. See section argument.
    -t, --timeout <time>    Auto exit. [default: 30]

commands:
    line                    Draw a line.
    circle                  Draw a circle.
    ellipse                 Draw a ellipse.
    fill                    Fill a region.
    bezier                  Draw a Bezier curve.
    b                       Draw a B-spline.
    koch                    Draw a Koch curve.
    mandelbrot              Draw a Mandelbrot set.
    julia                   Draw a Julia set.
    fern                    Draw a fern.
    reality                 Draw a scene by hidden surface removal, mirror
                            reflection, texture, etc.

output: (%d is format string)
    *%d*.png
    *%d*.jpg
    *%d*.bmp
    */video.mp4
    */video.gif

algorithm:
                default                     second
    line        midpoint                    bresenham
    circle      midpoint                    bresenham
    ellipse     midpoint
    fill        scan                        seed
    bezier      Bezier
    b           De_Boor
    koch        Lindenmayer_system
    mandelbrot  f(z)=z*z+c
    julia       f(z)=z*z+c
    fern        iterated_fuction_system
    reality     ray_tracing

argument:
    line        x1,y1,x2,y2,...
    circle      x0,y0,r
    ellipse     x0,y0,a,b
    fill (scan) x1,y1,x2,y2,x3,y3,...
    fill (seed) x_seed,y_seed,x1,y1,x2,y2,x3,y3,...
    bezier      sample_number,x1,y1,...
    b           degree,x1,y1,...
    koch        scale,iteration_max,x0,y0
    mandelbrot  scale,iteration_max
    julia       scale,iteration_max
    fern        scale,iteration_max,x0,y0
    reality     sample_number,r_aperture

Report bugs to <wuzy01@qq.com>.
"""
# }}} Docs #
from typing import Optional, Dict, Union, Final
from pprint import pformat
import logging
from docopt import docopt
from . import __version__, _binname as BINNAME

VERSION: Final = """{version}
Copyright (C) 2022
Written by Wu Zhenyu
""".format(
    version=__version__
)
DOC: Final = __doc__.format(binname=BINNAME)  # type: ignore
Arg = Optional[Union[bool, int, str]]
logger = logging.getLogger(__name__)


def main(doc: str = DOC):
    """Run main function."""
    try:
        args: Dict[str, Arg] = docopt(doc, version=VERSION)
    except Exception:
        args = {}
    if args.get("--debug"):
        try:
            from rich.logging import RichHandler

            logging.basicConfig(
                level="DEBUG",
                format="%(message)s",
                handlers=[RichHandler(rich_tracebacks=True, markup=True)],
            )
        except ImportError:
            logging.basicConfig(level="DEBUG")
    logger.debug(pformat(args))
    # Parse {{{ #
    import os
    import taichi as ti

    if args["--quiet"]:
        import sys

        sys.stdout = open(os.devnull, "w")
        if args["--quiet"] > 1:  # type: ignore
            sys.stderr = sys.stdout

    R, G, B = list(map(int, args["--color"].split(",")))  # type: ignore
    if args["--args"]:
        argv = list(map(int, args["--args"].split(",")))  # type: ignore
    else:
        argv = []

    # debug=True will make print cannot work in ti.func
    ti.init(
        arch=ti.gpu,
        excepthook=True,
        advanced_optimization=(not args["--debug"]),
        verbose=args["--verbose"],
    )
    # }}} Parse #

    # pixels and paint must be defined in a same file
    if args["<command>"] in ["line"]:
        from .utils import init, pixels

        if args["--second"]:
            from .line.bresenham import paint
        else:
            from .line.midpoint import paint
        argv = argv if argv else [100, 5, 3, 3]
    elif args["<command>"] in ["circle"]:
        from .utils import init, pixels

        if args["--second"]:
            from .circle.bresenham import paint
        else:
            from .circle.midpoint import paint
        argv = argv if argv else [100, 50, 25]
    elif args["<command>"] in ["ellipse"]:
        from .utils import init, pixels
        from .ellipse import paint

        argv = argv if argv else [100, 50, 30, 10]
    elif args["<command>"] in ["fill"]:
        from .utils import init, pixels

        if args["--second"]:
            from .fill.seed import paint

            argv = argv if argv else [50, 30, 0, 0, 100, 50, 25, 100]
        else:
            from .fill.scan import paint

            argv = argv if argv else [0, 0, 100, 50, 50, 30, 25, 100]

    elif args["<command>"] in ["bezier"]:
        from .utils import init, pixels
        from .bezier import paint

        argv = argv if argv else [100, 0, 0, 30, 300, 300, 300]
    elif args["<command>"] in ["b"]:
        from .utils import init, pixels
        from .b import paint

        argv = (
            argv
            if argv
            else [
                3,
                0,
                0,
                100,
                300,
                200,
                100,
                300,
                300,
                400,
                100,
                500,
                400,
                600,
                0,
                700,
                500,
            ]
        )
    elif args["<command>"] in ["koch"]:
        from .utils import init, pixels
        from .koch import paint

        argv = argv if argv else [4, 5, 0, 0]
    elif args["<command>"] in ["mandelbrot"]:
        from .utils import init, pixels
        from .mandelbrot import paint

        argv = argv if argv else [2, 20]
    elif args["<command>"] in ["julia"]:
        from .utils import init, pixels
        from .julia import paint

        argv = argv if argv else [2, 20]
    elif args["<command>"] in ["fern"]:
        from .utils import init, pixels
        from .fern import paint

        argv = argv if argv else [50, 2000, 512, 0]
    elif args["<command>"] in ["reality"]:
        from .utils import pixels
        from .reality import paint, init, spheres

        ti.root.dense(ti.i, 20).place(spheres)
        argv = argv if argv else [4, 0.05]
    else:
        _name = os.path.split(__file__)[-1]
        exit(f"{args['<command>']} is an illegal command. See '{_name} -h'.")

    if args["--verbose"]:
        print(argv)

    out: Optional[str] = args["--output"]  # type: ignore
    if not out:
        out = ""
    ext = out.split(".")[-1]
    output_dir = os.path.split(out)[0]
    if output_dir == "":
        output_dir = "."
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    if ext in ["mp4", "gif"]:
        video_manager = ti.VideoManager(
            output_dir=output_dir, automatic_build=False
        )
    elif ext in ["png", "jpg", "bmp"]:
        i = 0

    shape = (int(args["--width"]), int(args["--height"]))  # type: ignore
    ti.root.dense(ti.ij, shape).place(pixels)
    init()
    gui = ti.GUI(args["<command>"], shape)
    import time

    timeout = float(args["--timeout"])  # type: ignore
    start = time.time()
    while time.time() - start < timeout and gui.running:
        paint(R, G, B, *argv)
        if not args["--dry-run"]:
            gui.set_image(pixels)
            gui.show()
        if ext in ["mp4", "gif"]:
            video_manager.write_frame(pixels)  # type: ignore
        elif ext in ["png", "jpg", "bmp"]:
            i += 1  # type: ignore
            try:
                file = out % i
            except TypeError:
                file = out
            ti.imwrite(pixels, file)

    if ext == "mp4":
        video_manager.make_video(gif=False)  # type: ignore
    if ext == "gif":
        video_manager.make_video(mp4=False)  # type: ignore


if __name__ == "__main__":
    main()
# ex: foldmethod=marker



"""``computer_graphics_demo/__init__.py``."""

import logging
from typing import Final

logger = logging.getLogger(__name__)

try:
    from get_version import get_version, NoVersionFound

    try:
        __version__ = get_version(__file__)
    except NoVersionFound:
        __version__ = "0.0.0"
except ImportError:
    __version__ = "Please install get_version!"

_binname: Final = "cgdemo"



"""
computer_graphics_demo.ellipse
==============================

Draw a ellipse.
"""
import taichi as ti
from .circle import set4pixel


@ti.func
def ellipse(R: int, G: int, B: int, x0: int, y0: int, a0: int, b0: int):
    """ellipse.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param x0:
    :type x0: int
    :param y0:
    :type y0: int
    :param a:
    :type a: int
    :param b:
    :type b: int
    """
    a, b = ti.min(a0, b0), ti.max(a0, b0)
    x, y, p = 0, b, b ** 2 - b * a ** 2 - a ** 2 / 4

    while x * b ** 2 <= y * a ** 2:
        if p < 0:
            p += 2 * x * b ** 2 + 3 * b ** 2
        else:
            p += 2 * x ** b * 2 - 2 * y * a ** 2 + 3 * b ** 2 + 2 * a ** 2
            y -= 1
        if a == a0:
            set4pixel(x0, y0, x, y, R, G, B)
        else:
            set4pixel(x0, y0, y, x, R, G, B)
        x += 1
    q = (b * (x + 1 / 2)) ** 2 + (a * (y - 1)) ** 2 - (a * b) ** 2
    while y >= 0:
        if q < 0:
            q += 3 * a ** 2 + 2 * b ** 2 + 2 * x * b ** 2 - 2 * y * a ** 2
            x += 1
        else:
            q += 3 * a ** 2 - 2 * y * a ** 2
        if a == a0:
            set4pixel(x0, y0, x, y, R, G, B)
        else:
            set4pixel(x0, y0, y, x, R, G, B)
        y -= 1


@ti.kernel
def paint(R: int, G: int, B: int, x0: int, y0: int, a: int, b: int):
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param x0:
    :type x0: int
    :param y0:
    :type y0: int
    :param a:
    :type a: int
    :param b:
    :type b: int
    """
    ellipse(R, G, B, x0, y0, a, b)



"""
computer_graphics_demo.fill.seed
================================

Seed point algorithm.
"""
from ..line import polygon
from ..line.bresenham import line
from ..utils import pixels, setpixel


def paint(R: int, G: int, B: int, x_seed: int, y_seed: int, *argv: int):
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param x_seed:
    :type x_seed: int
    :param y_seed:
    :type y_seed: int
    :param argv:
    :type argv: int
    """
    if len(argv) % 2:
        exit(f"point number must be even, now is {len(argv)}")
    elif len(argv) < 6:
        exit(f"point number must be greater than 3, now is {len(argv)}")
    polygon(R, G, B, *argv, isclose=True, line=line)

    stack = [(x_seed, y_seed)]
    while len(stack):
        x, y = stack[0]
        setpixel(x, y, R, G, B)
        stack.remove((x, y))
        for i, j in [[0, 1], [0, -1], [1, 0], [-1, 0]]:
            x2, y2 = x + i, y + j
            # pixels[x2, y2] have already been colorified
            if (
                pixels[x2, y2][0]
                or pixels[x2, y2][1]
                or pixels[x2, y2][2]
                or set(stack) | {(x2, y2)} == set(stack)
            ):
                continue
            stack += [(x2, y2)]



"""
computer_graphics_demo.circle.midpoint
======================================

Midpoint algorithm.
"""
import taichi as ti
from . import set8pixel


@ti.func
def circle(R: int, G: int, B: int, x0: int, y0: int, r: int):
    """circle.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param x0:
    :type x0: int
    :param y0:
    :type y0: int
    :param r:
    :type r: int
    """
    x, y = 0, r

    p = 5 / 4 - r
    while x <= y:
        if p < 0:
            p += 2 * x + 3
        else:
            p += 2 * x - 2 * y + 5
            y -= 1
        set8pixel(x0, y0, x, y, R, G, B)
        x += 1


@ti.kernel
def paint(R: int, G: int, B: int, x0: int, y0: int, r: int):
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param x0:
    :type x0: int
    :param y0:
    :type y0: int
    :param r:
    :type r: int
    """
    circle(R, G, B, x0, y0, r)



"""
computer_graphics_demo
======================

Common functions.
"""

import taichi as ti
import logging

logger = logging.getLogger(__name__)

pixels = ti.Vector.field(3, float)


def init():
    """Init."""
    pass


@ti.pyfunc
def setpixel(x: int, y: int, R: int, G: int, B: int):
    """Set Pixel.

    :param x:
    :type x: int
    :param y:
    :type y: int
    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    """
    pixels[x, y] = ti.Vector([R, G, B])



"""
computer_graphics_demo.line.midpoint
====================================

Midpoint algorithm.
"""
import taichi as ti
from ..utils import setpixel
from . import polygon


@ti.pyfunc
def line(R: int, G: int, B: int, x1: int, y1: int, x2: int, y2: int):
    """line.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param x1:
    :type x1: int
    :param y1:
    :type y1: int
    :param x2:
    :type x2: int
    :param y2:
    :type y2: int
    """
    m = 100
    if x2 != x1:
        m = (y2 - y1) / (x2 - x1)

    u1, u2, v1, v2 = x1, x2, y1, y2
    if ti.abs(m) > 1:
        u1, u2, v1, v2 = y1, y2, x1, x2
    if u1 > u2:
        u1, u2, v1, v2 = u2, u1, v2, v1
    a, b = ti.abs(v2 - v1), u1 - u2
    p = a + 0.5 * b
    u, v = u1, v1
    while u < u2:
        if p < 0:
            p += a
        else:
            p += a + b
            if m > 0:
                v += 1
            else:
                v -= 1
        if ti.abs(m) <= 1:
            setpixel(u, v, R, G, B)
        else:
            setpixel(v, u, R, G, B)
        u += 1


def paint(R: int, G: int, B: int, *argv: int) -> None:
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param argv:
    :type argv: int
    :rtype: None
    """
    polygon(R, G, B, *argv, line=line)



#! /usr/bin/env python
"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

from glob import glob
from mimetypes import guess_type
import os
from pathlib import Path
from typing import Final

from get_version import get_version
from packaging.utils import canonicalize_name
from setuptools import find_packages, setup

here = Path(__file__).parent.resolve()
try:
    readme = next(iter(here.glob("README*")))
    long_description = readme.read_text()
    long_description_content_type = guess_type(readme)[0]
except StopIteration:
    long_description = ""
    long_description_content_type = None
REQUIREMENTS: Final = here / "requirements.txt"
try:
    install_requires = REQUIREMENTS.read_text().splitlines()
except FileNotFoundError:
    install_requires = []
extras_require = {}
for file in glob("requirements/*.txt"):
    name = os.path.basename(file)[:-4]
    extras_require[name] = Path(file).read_text().splitlines()
extras_require["all"] = sum(extras_require.values(), [])

VERSION: Final = get_version(__file__)
BINNAME: Final = "cgdemo"
NAME_: Final = "computer_graphics_demo"
NAME: Final = canonicalize_name(NAME_)
VCS_URL: Final = "https://github.com/Freed-Wu/computer-graphics-homework"

setup(
    name=NAME,
    version=VERSION,
    description="A demonstration of computer graphics",
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    url=f"https://pypi.org/project/{NAME}",
    download_url=f"{VCS_URL}/tags",
    project_urls={
        "Bug Reports": f"{VCS_URL}/issues",
        "Source": VCS_URL,
    },
    author="wzy",
    author_email="wuzy01@qq.com",
    maintainer="wzy",
    maintainer_email="wuzy01@qq.com",
    license="GPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
    ]
    + [
        f"Programming Language :: Python :: 3.{minor}"
        for minor in range(6, 10)
    ],
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        "console_scripts": [
            f"{BINNAME}={NAME_}.__main__:main",
        ],
    },
    keywords="python computer graphics",
)



"""
computer_graphics_demo.mandelbrot
=================================

Draw a Mandelbrot set
"""
import taichi as ti
from .utils import pixels, setpixel

n = 512


@ti.func
def complex_sqr(z) -> ti.Vector:
    """complex_sqr.

    :param z:
    """
    return ti.Vector([z[0] ** 2 - z[1] ** 2, z[1] * z[0] * 2])


@ti.kernel
def paint(R: int, G: int, B: int, scale: float, iteration_max: int):
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param scale:
    :type scale: float
    :param iteration_max:
    :type iteration_max: int
    """
    for i, j in pixels:
        c = ti.Vector([i / n - 1, j / n - 0.5]) * scale
        iterations = 0
        z = ti.Vector([0.0, 0.0])
        while z.norm() < 20 and iterations <= iteration_max:
            z = complex_sqr(z) + c
            iterations += 1
        if iterations > iteration_max:
            setpixel(i, j, R, G, B)



"""
computer_graphics_demo.circle
=============================

Draw a circle.
"""
from ..utils import setpixel


def set4pixel(x0, y0, x, y, R, G, B):
    """set4pixel.

    :param x0:
    :param y0:
    :param x:
    :param y:
    :param R:
    :param G:
    :param B:
    """
    setpixel(x0 + x, y0 + y, R, G, B)
    setpixel(x0 - x, y0 + y, R, G, B)
    setpixel(x0 + x, y0 - y, R, G, B)
    setpixel(x0 - x, y0 - y, R, G, B)


def set8pixel(x0, y0, x, y, R, G, B):
    """set8pixel.

    :param x0:
    :param y0:
    :param x:
    :param y:
    :param R:
    :param G:
    :param B:
    """
    set4pixel(x0, y0, x, y, R, G, B)
    set4pixel(x0, y0, y, x, R, G, B)



"""
computer_graphics_demo.line
===========================

Draw a line.
"""
from ..fill import cycle_shift
from ..utils import setpixel
from typing import Callable, Optional


def polygon(
    R: int,
    G: int,
    B: int,
    *argv: int,
    isclose: bool = False,
    line: Optional[Callable[[int, int, int, int, int, int, int], None]] = None,
) -> None:
    """polygon.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param argv:
    :type argv: int
    :param isclose:
    :type isclose: bool
    :param line:
    :type line: Optional[Callable[[int, int, int, int, int, int, int], None]]
    :rtype: None
    """
    xs = argv[::2]
    ys = argv[1::2]
    if len(argv) % 2:
        exit(f"point number must be even, now is {len(argv)}")
    elif len(argv) % 2:
        exit(f"point number must be greater than 2, now is {len(xs)}")
    vertexs = list(zip(xs, ys))
    vertexs0 = cycle_shift(vertexs, -1)
    if isclose:
        edges = zip(vertexs0, vertexs)
    else:
        edges = zip(vertexs0[1:], vertexs[1:])
    for (x1, y1), (x2, y2) in edges:
        if line:
            line(R, G, B, x1, y1, x2, y2)
        else:
            setpixel(R, G, B, x1, y1)
            setpixel(R, G, B, x2, y2)
